#pragma once

#include "../../Global.h"

// This header should be treated as a pre compiled header file(PCH) in the compiler building settings.

// Here you can add specific libraries and macros which are specific to this project.

